package net.knarcraft.blacksmith.command.blacksmith;

import net.knarcraft.blacksmith.config.SettingValueType;
import net.knarcraft.blacksmith.config.blacksmith.BlacksmithSetting;
import net.knarcraft.blacksmith.util.InputParsingHelper;
import net.knarcraft.knarlib.util.TabCompletionHelper;
import org.bukkit.command.Command;
import org.bukkit.command.CommandSender;
import org.bukkit.command.TabCompleter;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.ArrayList;
import java.util.List;

import static net.knarcraft.blacksmith.util.TabCompleteValuesHelper.getTabCompletions;
import static net.knarcraft.knarlib.util.TabCompletionHelper.filterMatchingContains;

/**
 * The tab completer for the command used for changing global blacksmith configuration options
 */
public class BlackSmithConfigTabCompleter implements TabCompleter {

    @Nullable
    @Override
    public List<String> onTabComplete(@NotNull CommandSender sender, @NotNull Command command, @NotNull String label,
                                      @NotNull String[] args) {
        if (!sender.hasPermission("blacksmith.admin")) {
            return new ArrayList<>();
        }

        //Arguments: <setting> [new value/material or enchantment] []

        //Prevent tab-completion when typing messages with spaces
        if (args.length > 2 && skipCompletionForSpacedMessage(args[0]) != null) {
            return new ArrayList<>();
        }

        if (args.length == 1) {
            List<String> availableCommands = new ArrayList<>();
            availableCommands.add("reload");
            for (BlacksmithSetting setting : BlacksmithSetting.values()) {
                availableCommands.add(setting.getCommandName());
            }
            return filterMatchingContains(availableCommands, args[0]);
        } else if (args.length == 2) {
            return tabCompleteCommandValues(args[0], args[1]);
        } else if (args.length == 3) {
            //Get per-material tab completions, or return nothing if an invalid setting was specified
            BlacksmithSetting blacksmithSetting = BlacksmithSetting.getSetting(args[0]);
            if (blacksmithSetting != null) {
                return getPerTypeTabCompletions(blacksmithSetting, args);
            } else {
                return new ArrayList<>();
            }
        }
        return new ArrayList<>();
    }

    /**
     * Checks whether to tab-complete nothing because a message containing spaces is being written
     *
     * @param command <p>The command specified by the user</p>
     * @return <p>Null if not writing a spaced message</p>
     */
    private List<String> skipCompletionForSpacedMessage(@NotNull String command) {
        BlacksmithSetting blacksmithSetting = BlacksmithSetting.getSetting(command);
        if (blacksmithSetting != null && blacksmithSetting.isMessage()) {
            return new ArrayList<>();
        }
        return null;
    }

    /**
     * Gets tab-completions for a selected material or enchantment
     *
     * @param blacksmithSetting <p>The global setting to get tab-completions for</p>
     * @param args              <p>The arguments given by the user</p>
     * @return <p>The tab-completions to show to the user</p>
     */
    private List<String> getPerTypeTabCompletions(@NotNull BlacksmithSetting blacksmithSetting,
                                                  @NotNull String[] args) {
        //Display possible tab-completions only if a valid enchantment or material is provided
        if (((blacksmithSetting == BlacksmithSetting.BASE_PRICE ||
                blacksmithSetting == BlacksmithSetting.PRICE_PER_DURABILITY_POINT) &&
                InputParsingHelper.matchMaterial(args[1]) != null) ||
                (blacksmithSetting == BlacksmithSetting.ENCHANTMENT_COST &&
                        InputParsingHelper.matchEnchantment(args[1]) != null)) {
            return filterMatchingContains(getTabCompletions(blacksmithSetting.getValueType()), args[2]);
        } else {
            return new ArrayList<>();
        }
    }

    /**
     * Tab completes the values available for the given command
     *
     * @param commandName  <p>The name of the used command</p>
     * @param commandValue <p>The command value used to filter tab-completions</p>
     * @return <p>Some valid options for the command's argument</p>
     */
    private List<String> tabCompleteCommandValues(@NotNull String commandName, @NotNull String commandValue) {
        if (commandName.equalsIgnoreCase("reload")) {
            return new ArrayList<>();
        }
        BlacksmithSetting setting = BlacksmithSetting.getSetting(commandName);
        if (setting != null) {
            return getCompletions(setting, commandValue);
        } else {
            return null;
        }
    }

    /**
     * Gets tab-completions for the given global setting and filters on the command value
     *
     * @param blacksmithSetting <p>The global setting to get tab-completions for</p>
     * @param commandValue      <p>The command value used to filter between available tab-completions</p>
     * @return <p>The available tab-completions</p>
     */
    private List<String> getCompletions(@NotNull BlacksmithSetting blacksmithSetting,
                                        @NotNull String commandValue) {
        if (blacksmithSetting == BlacksmithSetting.ENCHANTMENT_BLOCK_LIST ||
                blacksmithSetting == BlacksmithSetting.REFORGE_ABLE_ITEMS) {
            return TabCompletionHelper.getStringList(getTabCompletions(blacksmithSetting.getValueType()),
                    commandValue, TabCompletionHelper::filterMatchingContains);
        }

        List<String> returnValues = filterMatchingContains(
                getTabCompletions(blacksmithSetting.getValueType()), commandValue);
        if (blacksmithSetting == BlacksmithSetting.BASE_PRICE ||
                blacksmithSetting == BlacksmithSetting.PRICE_PER_DURABILITY_POINT) {
            returnValues.addAll(filterMatchingContains(getTabCompletions(SettingValueType.MATERIAL), commandValue));
        } else if (blacksmithSetting == BlacksmithSetting.ENCHANTMENT_COST) {
            returnValues.addAll(filterMatchingContains(getTabCompletions(SettingValueType.ENCHANTMENT), commandValue));
        }
        return returnValues;
    }

}
