package net.knarcraft.blacksmith.config.blacksmith;

import net.citizensnpcs.api.util.DataKey;
import net.knarcraft.blacksmith.BlacksmithPlugin;
import net.knarcraft.blacksmith.config.SettingValueType;
import net.knarcraft.blacksmith.config.TraitSettings;
import net.knarcraft.blacksmith.util.ConfigHelper;
import net.knarcraft.blacksmith.util.InputParsingHelper;
import net.knarcraft.blacksmith.util.ItemHelper;
import org.bukkit.Material;
import org.bukkit.enchantments.Enchantment;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * A class which keeps track of all Blacksmith settings/config values for one NPC
 */
public class BlacksmithNPCSettings implements TraitSettings<BlacksmithSetting> {

    private final List<Material> reforgeAbleItems = new ArrayList<>();
    private final List<Enchantment> enchantmentBlockList = new ArrayList<>();
    private final Map<BlacksmithSetting, Object> currentValues = new HashMap<>();
    private final GlobalBlacksmithSettings globalBlacksmithSettings;

    /**
     * Instantiates a new "Settings" object
     */
    public BlacksmithNPCSettings(GlobalBlacksmithSettings globalBlacksmithSettings) {
        this.globalBlacksmithSettings = globalBlacksmithSettings;
    }

    /**
     * Loads variables from the given data key
     *
     * @param key <p>The data key to load variables from</p>
     */
    public void loadVariables(DataKey key) {
        for (BlacksmithSetting setting : BlacksmithSetting.values()) {
            if (key.keyExists(setting.getChildPath())) {
                currentValues.put(setting, key.getRaw(setting.getChildPath()));
            }
        }
        //Updates the list of reforge-able items/materials
        updateReforgeAbleItems();
        updateEnchantmentBlockList();
    }

    /**
     * Saves variables to the given data key
     *
     * @param key <p>The data key to save variables to</p>
     */
    public void saveVariables(DataKey key) {
        for (BlacksmithSetting setting : BlacksmithSetting.values()) {
            key.setRaw(setting.getChildPath(), currentValues.get(setting));
        }
    }

    @Override
    public void changeValue(@NotNull BlacksmithSetting setting, @Nullable Object newValue) {
        if (setting.getValueType() == SettingValueType.ENCHANTMENT_LIST ||
                setting.getValueType() == SettingValueType.REFORGE_ABLE_ITEMS) {
            //Workaround to make sure it's treated as the correct type
            currentValues.put(setting, newValue == null ? null : ConfigHelper.asStringList(newValue));
        } else {
            currentValues.put(setting, newValue);
        }
        if (setting == BlacksmithSetting.REFORGE_ABLE_ITEMS) {
            updateReforgeAbleItems();
        }
        if (setting == BlacksmithSetting.ENCHANTMENT_BLOCK_LIST) {
            updateEnchantmentBlockList();
        }
    }

    /**
     * Gets the raw current value of a setting
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The current value of the setting</p>
     */
    public @NotNull Object getRawValue(@NotNull BlacksmithSetting setting) {
        return currentValues.get(setting);
    }

    @Override
    public String getBusyWithPlayerMessage() {
        return asString(BlacksmithSetting.BUSY_WITH_PLAYER_MESSAGE);
    }

    /**
     * Gets the message to display when the blacksmith is busy with reforging an item
     *
     * @return <p>The busy reforging message</p>
     */
    public String getBusyWorkingMessage() {
        return asString(BlacksmithSetting.BUSY_WITH_REFORGE_MESSAGE);
    }

    /**
     * Gets the message to use for displaying an item's cost
     *
     * @return <p>The message to use for displaying item cost</p>
     */
    public String getCostMessage() {
        return asString(BlacksmithSetting.COST_MESSAGE);
    }

    /**
     * Gets the message to display when a blacksmith has been given an invalid item
     *
     * @return <p>The invalid item message</p>
     */
    public String getInvalidItemMessage() {
        return asString(BlacksmithSetting.INVALID_ITEM_MESSAGE);
    }

    /**
     * Gets the message to display when the presented item is at full durability
     *
     * @return <p>The not damaged message</p>
     */
    public String getNotDamagedMessage() {
        return asString(BlacksmithSetting.NOT_DAMAGED_MESSAGE);
    }

    @Override
    public String getStartWorkingMessage() {
        return asString(BlacksmithSetting.START_REFORGE_MESSAGE);
    }

    /**
     * Gets the message to display when a blacksmith has successfully repaired an item
     *
     * @return <p>The reforge success message</p>
     */
    public String getSuccessMessage() {
        return asString(BlacksmithSetting.SUCCESS_MESSAGE);
    }

    /**
     * Gets the message to display when a blacksmith has failed to repair an item
     *
     * @return <p>The reforge fail message</p>
     */
    public String getFailMessage() {
        return asString(BlacksmithSetting.FAIL_MESSAGE);
    }

    /**
     * Gets the message to display when a player cannot afford re-forging an item
     *
     * @return <p>The insufficient funds message</p>
     */
    public String getInsufficientFundsMessage() {
        return asString(BlacksmithSetting.INSUFFICIENT_FUNDS_MESSAGE);
    }

    @Override
    public String getCoolDownUnexpiredMessage() {
        return asString(BlacksmithSetting.COOL_DOWN_UNEXPIRED_MESSAGE);
    }

    /**
     * Gets the message to display when a player has changed the item they're trying to reforge
     *
     * @return <p>The item changed message</p>
     */
    public String getItemChangedMessage() {
        return asString(BlacksmithSetting.ITEM_UNEXPECTEDLY_CHANGED_MESSAGE);
    }

    /**
     * Gets the message displayed if a player presents the blacksmith with an empty hand
     *
     * @return <p>The no item message</p>
     */
    @NotNull
    public String getNoItemMessage() {
        return asString(BlacksmithSetting.NO_ITEM_MESSAGE);
    }

    /**
     * Gets all items reforge-able by this NPC
     *
     * <p>If this is not empty, only the items specified can be repaired by this NPC.</p>
     *
     * @return <p>All items reforge-able by this NPC</p>
     */
    public List<Material> getReforgeAbleItems() {
        Object currentValue = currentValues.get(BlacksmithSetting.REFORGE_ABLE_ITEMS);
        if (currentValue == null || String.valueOf(currentValue).isEmpty()) {
            return globalBlacksmithSettings.getReforgeAbleItems();
        } else {
            return new ArrayList<>(this.reforgeAbleItems);
        }
    }

    /**
     * Gets the list of blocked enchantments
     *
     * @return <p>The list of blocked enchantments</p>
     */
    public List<Enchantment> getEnchantmentBlockList() {
        Object currentValue = currentValues.get(BlacksmithSetting.ENCHANTMENT_BLOCK_LIST);
        if (currentValue == null || String.valueOf(currentValue).isEmpty()) {
            return globalBlacksmithSettings.getEnchantmentBlockList();
        } else {
            return new ArrayList<>(this.enchantmentBlockList);
        }
    }

    /**
     * Gets the minimum delay used to wait for a reforging to finish.
     *
     * @return <p>The minimum reforge delay</p>
     */
    public int getMinReforgeDelay() {
        return asInt(BlacksmithSetting.MIN_REFORGE_DELAY);
    }

    /**
     * Gets the maximum delay used to wait for a reforging to finish
     *
     * @return <p>The maximum reforge delay</p>
     */
    public int getMaxReforgeDelay() {
        return asInt(BlacksmithSetting.MAX_REFORGE_DELAY);
    }

    /**
     * Gets the cool-down between each reforge
     *
     * @return <p>The reforge cool-down</p>
     */
    public int getReforgeCoolDown() {
        return asInt(BlacksmithSetting.REFORGE_COOL_DOWN);
    }

    /**
     * Gets the chance to fail a reforging
     *
     * @return <p>The fail chance</p>
     */
    public int getFailChance() {
        return asInt(BlacksmithSetting.FAIL_CHANCE);
    }

    /**
     * Gets whether a failed reforging should remove/downgrade enchantments
     *
     * @return <p>Whether enchantments should be removed</p>
     */
    public boolean getFailRemovesEnchantments() {
        return asBoolean(BlacksmithSetting.FAIL_REMOVE_ENCHANTMENTS);
    }

    /**
     * Gets the chance for adding an extra enchantment to an item
     *
     * @return <p>The extra enchantment chance</p>
     */
    public int getExtraEnchantmentChance() {
        return asInt(BlacksmithSetting.EXTRA_ENCHANTMENT_CHANCE);
    }

    /**
     * Gets the max number of enchantment to add to an item
     *
     * @return <p>The maximum enchantments</p>
     */
    public int getMaxEnchantments() {
        return asInt(BlacksmithSetting.MAX_ENCHANTMENTS);
    }

    /**
     * Gets whether an item should be dropped on the ground instead of being given to the player
     *
     * @return <p>Whether to drop reforged items on the ground</p>
     */
    public boolean getDropItem() {
        return asBoolean(BlacksmithSetting.DROP_ITEM);
    }

    /**
     * Gets the title of the blacksmith, describing which items it can reforge
     *
     * @return <p>The title of the blacksmith</p>
     */
    public String getBlacksmithTitle() {
        return asString(BlacksmithSetting.BLACKSMITH_TITLE);
    }

    @Override
    public boolean getDisableCoolDown() {
        return asInt(BlacksmithSetting.REFORGE_COOL_DOWN) <= 0;
    }

    /**
     * Gets whether this blacksmith is able to repair anvils
     *
     * @return <p>True if this blacksmith is able to repair anvils</p>
     */
    public boolean getRepairAnvils() {
        return asBoolean(BlacksmithSetting.REPAIR_ANVILS);
    }

    /**
     * Gets the given value as an integer
     *
     * <p>This will throw an exception if used for a non-integer setting</p>
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The value of the given setting as an integer</p>
     */
    private int asInt(BlacksmithSetting setting) {
        return ConfigHelper.asInt(getValue(setting));
    }

    /**
     * Gets the string value of the given setting
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The value of the given setting as a string</p>
     */
    private String asString(BlacksmithSetting setting) {
        return getValue(setting).toString();
    }

    /**
     * Gets the boolean value of the given setting
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The value of the given setting as a boolean</p>
     */
    private boolean asBoolean(BlacksmithSetting setting) {
        return ConfigHelper.asBoolean(getValue(setting));
    }

    /**
     * Gets the value of a setting, using the default if not set
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The current value</p>
     */
    private Object getValue(BlacksmithSetting setting) {
        Object value = currentValues.get(setting);
        //If not set, use the default value from the config.yml file
        if (value == null) {
            Map<BlacksmithSetting, Object> defaultNPCSettings = globalBlacksmithSettings.getDefaultNPCSettings();
            if (defaultNPCSettings.containsKey(setting)) {
                value = defaultNPCSettings.get(setting);
            }
        }
        //If not set in config.yml, use the default value from the enum
        if (value == null) {
            value = setting.getDefaultValue();
        }
        return value;
    }

    /**
     * Updates the list of blocked enchantments
     */
    private void updateEnchantmentBlockList() {
        this.enchantmentBlockList.clear();
        List<String> enchantments = ConfigHelper.asStringList(getValue(BlacksmithSetting.ENCHANTMENT_BLOCK_LIST));
        if (enchantments != null) {
            this.enchantmentBlockList.addAll(getEnchantmentBlockList(enchantments));
        }
    }

    /**
     * Gets the list of enchantments listed in the given string list
     *
     * @param enchantments <p>The enchantment names to block</p>
     * @return <p>The enchantments to be blocked</p>
     */
    public static List<Enchantment> getEnchantmentBlockList(List<String> enchantments) {
        List<Enchantment> enchantmentBlockList = new ArrayList<>();

        for (String item : enchantments) {
            if (InputParsingHelper.isEmpty(item)) {
                continue;
            }

            Enchantment enchantment = InputParsingHelper.matchEnchantment(item);
            if (enchantment != null) {
                enchantmentBlockList.add(enchantment);
            } else {
                BlacksmithPlugin.warn("Unable to verify " + item + " as a valid enchantment");
            }
        }
        return enchantmentBlockList;
    }

    /**
     * Updates the reforge-able items according to the current value of the setting
     */
    private void updateReforgeAbleItems() {
        this.reforgeAbleItems.clear();
        List<String> materialStrings = ConfigHelper.asStringList(getValue(BlacksmithSetting.REFORGE_ABLE_ITEMS));
        if (materialStrings != null) {
            this.reforgeAbleItems.addAll(ItemHelper.getItems(materialStrings, true));
        }
    }

}