package net.knarcraft.blacksmith.trait;

import net.citizensnpcs.api.util.DataKey;
import net.knarcraft.blacksmith.BlacksmithPlugin;
import net.knarcraft.blacksmith.config.blacksmith.BlacksmithNPCSettings;
import net.knarcraft.blacksmith.config.blacksmith.BlacksmithSetting;
import net.knarcraft.blacksmith.manager.EconomyManager;
import net.knarcraft.blacksmith.util.ItemHelper;
import net.knarcraft.knarlib.formatting.StringFormatter;
import org.bukkit.Bukkit;
import org.bukkit.Material;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.jetbrains.annotations.NotNull;

import java.util.List;

import static net.knarcraft.blacksmith.formatting.BlacksmithStringFormatter.sendNPCMessage;

/**
 * The class representing the Blacksmith NPC trait
 */
public class BlacksmithTrait extends CustomTrait<BlacksmithSetting> {

    private final BlacksmithNPCSettings config;

    /**
     * Instantiates a new blacksmith trait
     */
    public BlacksmithTrait() {
        super("blacksmith");
        //This should crash if the blacksmith plugin hasn't been properly registered
        Bukkit.getServer().getPluginManager().getPlugin("Blacksmith");
        this.config = new BlacksmithNPCSettings(BlacksmithPlugin.getInstance().getGlobalBlacksmithSettings());
        super.setTraitSettings(this.config);
    }

    /**
     * Gets the current settings for this NPC
     *
     * @return <p>The current settings for this NPC</p>
     */
    @NotNull
    public BlacksmithNPCSettings getSettings() {
        return config;
    }

    /**
     * Loads all config values stored in citizens' config file for this NPC
     *
     * @param key <p>The data key used for the config root</p>
     */
    @Override
    public void load(@NotNull DataKey key) {
        config.loadVariables(key);
    }

    /**
     * Saves all config values for this NPC
     *
     * @param key <p>The data key used for the config root</p>
     */
    @Override
    public void save(@NotNull DataKey key) {
        config.saveVariables(key);
    }

    @Override
    public void startSession(@NotNull Player player) {
        ItemStack hand = player.getInventory().getItemInMainHand();
        if (hand.getType().isAir()) {
            sendNPCMessage(this.npc, player, config.getNoItemMessage());
            return;
        }

        //Refuse if not repairable, or if reforge-able items is set, but doesn't include the held item
        List<Material> reforgeAbleItems = config.getReforgeAbleItems();

        boolean notHoldingAnvil = !this.config.getRepairAnvils() || !ItemHelper.isAnvil(hand.getType(), false);
        boolean notHoldingRepairable = !ItemHelper.isRepairable(hand) ||
                (!reforgeAbleItems.isEmpty() && !reforgeAbleItems.contains(hand.getType()));

        if (notHoldingAnvil && notHoldingRepairable) {
            String invalidMessage = StringFormatter.replacePlaceholder(config.getInvalidItemMessage(),
                    "{title}", config.getBlacksmithTitle());
            sendNPCMessage(this.npc, player, invalidMessage);
            return;
        }

        if (ItemHelper.getDamage(hand) == 0 && !ItemHelper.isAnvil(hand.getType(), true)) {
            sendNPCMessage(this.npc, player, config.getNotDamagedMessage());
            return;
        }

        //Start a new reforge session for the player
        currentSessionStartTime = System.currentTimeMillis();
        try {
            session = new ReforgeSession(this, player, npc, config);
        } catch (IllegalArgumentException exception) {
            BlacksmithPlugin.error(exception.getMessage());
            return;
        }

        //Tell the player the cost of repairing the item
        String cost = EconomyManager.formatBlacksmithCost(player);
        String itemName = hand.getType().name().toLowerCase().replace('_', ' ');
        sendNPCMessage(this.npc, player, config.getCostMessage().replace("{cost}", cost).replace("{item}",
                itemName));
    }

    @Override
    protected boolean showExactTime() {
        return BlacksmithPlugin.getInstance().getGlobalBlacksmithSettings().getShowExactTime();
    }

}
